
#ifndef __HIMOGL_H__
#define __HIMOGL_H__

#include <gl/glut.h>
#include <vector>
#include "matrix.h"
#include <windows.h>
#include <stdio.h>
#include "texture.h"

#define for if(0);else for

typedef Vector3<double> Vector3d;
typedef std::vector<Vector3d> Vector3dv;
typedef std::vector<double> doublev;
const double PI=3.14159265358979;
const int NN=20;
const double DT=0.001/NN;
const double DISTANCE=0.50;
const double COLLISION=0.58;
const double ACCEPTABLE=COLLISION*1.2;
const double ATTENUATION1=0.90;//΂˂Ƃ̂
const double ATTENUATION2=0.998;//CRIȂ

//Ɠ_̋
//p:̓_
//v:̐ڐxNg
//c:_
//v0łȂ
double CalcLinePointDistance(const Vector3d &p,const Vector3d &v,const Vector3d &c)
{
	Vector3d d=c;
	d-=p;
	double ab2=v.Abs2();
	// n/m = (p2-p1).(c-p1) / |p2-p1|^2
	double npm=v.DotProd(d)/(ab2);
	// d = p + (n/m) * (p2-p1) - c
	d=v;
	d*=npm;
	d+=p;
	d-=c;
	return d.Abs();
}

//_璼ɍ~鐂
//d:(cʂ鐂)
void CalcParpendicular(const Vector3d &p,const Vector3d &v,const Vector3d &c,Vector3d &d)
{
	d=c;
	d-=p;
	double ab2=v.Abs2();
	// n/m = (p2-p1).(c-p1) / |p2-p1|^2
	double npm=v.DotProd(d)/(ab2);
	// d = p + (n/m) * (p2-p1) - c
	d=v;
	d*=npm;
	d+=p;
	d-=c;
}

class HimoPart
{
public:
	Vector3d m_point;
	Vector3d m_oldpoint;
	Vector3d m_velocity;
	double collideradius;
};

typedef std::vector<HimoPart> HimoPartv;

class Himo
{
private:
	enum{POINTS=128};
	Vector3dv m_points;
	Vector3dv m_oldpoints;
	Vector3dv m_velocities;
	doublev m_collideradius;
	Vector3dv m_mainnormals,m_subnormals;

	int m_currentpoint;//ݒڂĂ_̃CfbNX
	bool m_iscaptured;//ЂړH
	Vector3d m_currentpointrecord;//ݒڂĂ_̃hbOJn̈ʒu
	Vector3d m_currentattracter;//hbÖʒu

	Texture m_floor;

private:

	//ii+1Ԃ̃xNg2̃m[}߂B
	void Normals(int i,Vector3d &b,Vector3d &n)const
	{
//		Vector3d t1=m_points[i]-m_points[i-1];
//		Vector3d t2=m_points[i+1]-m_points[i];
		Vector3d t=m_points[i+1]-m_points[i-1];
		b=t.ExProd(Vector3d(0,0,1));
		b.Normalize();
		n=b.ExProd(t);
		n.Normalize();
	}
	
	//΂
	void Spring(int direction)
	{
		int ini=0,fin=POINTS-1;
		if(direction==-1)
		{
			ini=POINTS-2;fin=-1;
		}
		
		//S
		double TotalLength=0;
		for(int i=0;i<POINTS-1;i++)
		{
			TotalLength+=m_points[i+1].Dist(m_points[i]);
		}

		if(m_iscaptured)
		{
			Vector3d t=m_currentattracter-m_points[m_currentpoint];
			double att=ATTENUATION1/2/( (TotalLength-POINTS*DISTANCE)*3+1 );
			t*=att;
			m_points[m_currentpoint]+=t;
		}
		
		//΂ˌvZ
		for(int i=ini;i!=fin;i+=direction)
		{
			Vector3d t=m_points[i+1]-m_points[i];
			double length=t.Abs();
			double displacement;
			if(i==0||i==POINTS-2)
			{
				if(TotalLength<POINTS*DISTANCE)
				{
					displacement = (DISTANCE - length) * ATTENUATION1/2;
				}
				else
				{
					displacement = (DISTANCE - length) * ATTENUATION1/2 / ( (TotalLength-POINTS*DISTANCE)*3+1 );
				}
			}
			else
			{
				displacement = (DISTANCE - length) * ATTENUATION1/2;
			}
			t*=(displacement/length);
			m_points[i]-=t;
			m_points[i+1]+=t;
		}
	}
	
	//蔻
	void Collision(int direction)
	{
		int ini=0,fin=POINTS-1;
		if(direction==-1)
		{
			ini=POINTS-2;fin=-1;
		}

		//^[Qbg_
		for(int i=ini;i!=fin;i+=direction)
		{
			//_m̏Փ
			for(int k=ini;k!=fin;k+=direction)
			{
				if(i!=k && i+1!=k && i-1!=k)//ׂ̐_dȂĂĂ܂B
				{
					Vector3d t=m_points[k]-m_points[i];
					double length=t.Abs();
					if(length<m_collideradius[k]+m_collideradius[i])//Փ
					{
						double d = (m_collideradius[k]+m_collideradius[i] - length) * ATTENUATION1/2;
						t*=(d/length);
						m_points[i]-=t;
						m_points[k]+=t;
					}
				}
			}

/*					//g[X
					//S:0,0,0
					{
					//a
					double r1=4,r2=1;
					
					//g[XƂ̋
					//sqrt( (sqrt(x^2+z^2)-r1)^2 + y^2) - r2

					double dis=sqrt( pow2(sqrt(pow2(p[j].x)+pow2(p[j].z))-r1) + pow2(p[j].y) )-r2;

					if(dis<c[j])//Փ
					{
						double ang=atan2(p[j].z,p[j].x);
						double dd=sqrt(pow2(p[j].x)+pow2(p[j].z))-r1;
						double ang2=atan2(p[j].y,dd);

						double rr=cos(ang2)*(r2+c[j])+r1;
						p[j].x=cos(ang)*rr;
						p[j].z=sin(ang)*rr;
						p[j].y=sin(ang2)*(r2+c[j]);

					}

					}

					//V_
					//y
					{
					//a
					double r=1;

					double dis=sqrt(pow2(p[j].x)+pow2(p[j].z))-r;

					if(dis<c[j])//Փ
					{
						double ang=atan2(p[j].z,p[j].x);
						double dd=sqrt(pow2(p[j].x)+pow2(p[j].z))-r;

						p[j].x=cos(ang)*(r+c[j]);
						p[j].z=sin(ang)*(r+c[j]);
					}

					}*/
		}
	}
	
public:
	void Init()
	{
		m_points.resize(POINTS);
		m_mainnormals.resize(POINTS);
		m_subnormals.resize(POINTS);
		m_oldpoints.resize(POINTS);
		m_velocities.resize(POINTS);
		m_collideradius.resize(POINTS);
		
		for(int i=0;i<POINTS;i++)
		{
			int j=i-(POINTS/2);
			m_points[i].x=j*0.2;
			m_points[i].y=0;
			m_points[i].z=0;
			
			m_velocities[i]=Vector3d(0,0,0);
			m_oldpoints[i]=m_points[i];
			m_collideradius[i]=COLLISION;
			
			m_mainnormals[i]=Vector3d(0,0,1);
			m_subnormals[i]=Vector3d(0,1,0);
		}
		m_oldpoints[0].z+=-5;
		m_oldpoints[POINTS-1].z-=-5;
		for(int i=1;i<POINTS-1;i++)
		{
			Normals(i,m_mainnormals[i],m_subnormals[i]);
		}
		m_currentpoint=-1;

		m_floor.LoadFile("tex_tile.png");
	}
	void Draw(const Vector3d &light)
	{
		const double h=-20-COLLISION-0.1;

		//
		m_floor.Activate();
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_DEPTH_WRITEMASK);
		glDisable(GL_LIGHTING);
		glColor3f(0.98,1.0,0.95);
		glBegin(GL_QUADS);
			glTexCoord2f(0,0);
			glVertex3f(40,-20-COLLISION,-40);
			glTexCoord2f(4,0);
			glVertex3f(-40,-20-COLLISION,-40);
			glTexCoord2f(4,4);
			glVertex3f(-40,-20-COLLISION,40);
			glTexCoord2f(0,4);
			glVertex3f(40,-20-COLLISION,40);
		glEnd();

		//ê悤
		glDisable(GL_TEXTURE_2D);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_DEPTH_WRITEMASK);
		glDisable(GL_LIGHTING);
		glColor4f(0,0,0,0.5);
		for(int i=1;i<POINTS-2;i++)
		{
			const int N=6;
			glBegin(GL_QUAD_STRIP);
			
			for(int j=0;j<=N;j++)
			{
				double s=sin(2*PI*j/N)*COLLISION,c=cos(2*PI*j/N)*COLLISION;
				Vector3d p=m_points[i],q=m_points[i+1];
				Vector3d n0=m_mainnormals[i],s0=m_subnormals[i];
				Vector3d n1=m_mainnormals[i+1],s1=m_subnormals[i+1];
				n0*=s;n1*=s;
				s0*=c;s1*=c;
				n0+=s0;n1+=s1;
				p+=n0;
				q+=n1;

				Vector3d psh=p-light;
				Vector3d qsh=q-light;

				if(psh.y>0)continue;
				if(qsh.y>0)continue;

				psh*=-(light.y-h)/psh.y;
				qsh*=-(light.y-h)/qsh.y;
				psh+=light;
				qsh+=light;



				glVertex3f(qsh.x,h,qsh.z);
				glVertex3f(psh.x,h,psh.z);
			}

			glEnd();
		}
		
		glEnable(GL_DEPTH_TEST);

		glEnable(GL_LIGHTING);
		for(int i=1;i<POINTS-2;i++)
		{
			const int N=10;
			glBegin(GL_QUAD_STRIP);

			for(int j=0;j<=N;j++)
			{
				double s=sin(2*PI*j/N)*COLLISION,c=cos(2*PI*j/N)*COLLISION;
				Vector3d p=m_points[i],q=m_points[i+1];
				Vector3d n0=m_mainnormals[i],s0=m_subnormals[i];
				Vector3d n1=m_mainnormals[i+1],s1=m_subnormals[i+1];
				n0*=s;n1*=s;
				s0*=c;s1*=c;
				n0+=s0;n1+=s1;
				p+=n0;
				q+=n1;

				glNormal3f(n1.x,n1.y,n1.z);
				glVertex3f(q.x,q.y,q.z);
				glNormal3f(n0.x,n0.y,n0.z);
				glVertex3f(p.x,p.y,p.z);
			}

			glEnd();
		}
		
		glColor4f(1,0,0,0.5);
		glDisable(GL_LIGHTING);
		glPointSize(5);
		glEnable(GL_POINT_SMOOTH);
		glBegin(GL_POINTS);
			glVertex3f(m_points[0].x,m_points[0].y,m_points[0].z);
			glVertex3f(m_points[POINTS-1].x,m_points[POINTS-1].y,m_points[POINTS-1].z);
		glEnd();
		glEnable(GL_LINE_SMOOTH);
		glBegin(GL_LINES);
			glVertex3f(m_points[0].x,m_points[0].y,m_points[0].z);
			glVertex3f(m_points[1].x,m_points[1].y,m_points[1].z);
			glVertex3f(m_points[POINTS-1].x,m_points[POINTS-1].y,m_points[POINTS-1].z);
			glVertex3f(m_points[POINTS-2].x,m_points[POINTS-2].y,m_points[POINTS-2].z);
		glEnd();

		if(m_currentpoint>=0)
		{
			glDisable(GL_DEPTH_TEST);
			glColor4f(0,0,1,0.5);
			glPointSize(10);
			glEnable(GL_POINT_SMOOTH);
			glBegin(GL_POINTS);
				glVertex3f(m_points[m_currentpoint].x,m_points[m_currentpoint].y,m_points[m_currentpoint].z);
			glEnd();
		}
	}


	void UpdatePhysicalSimulation()
	{
		static int direction=1;
		direction=-direction;
		
		static int t=0;
		t++;
		
/*		if(0x8000&GetAsyncKeyState(VK_SPACE))
		{
		m_oldpoints[0]=Vector3d(sin(t*0.005)*30,cos(t*0.05)*30,sin(t*0.05)*30);
		}
		else
		{
		m_oldpoints[0]=Vector3d(sin(t*0.005)*30,cos(t*0.05)*30,-sin(t*0.05)*30);
		}*/
		
		
		
		//ʒuXV
		for(int i=1;i<POINTS-1;i++)
		{
			m_points[i]+=m_velocities[i];
			m_points[i].y+=-0.98*DT*3;
		}

		//΂
		Spring(direction);
		
		//蔻
		Collision(-direction);

		for(int i=0;i<POINTS;i++)
		{
			if(i==0||i==POINTS-1)
			{
				m_points[i]=m_oldpoints[i];
				m_velocities[i]=Vector3d(0,0,0);
				continue;
			}
			int fg=0;

			if(m_points[i].y<-20)
			{
				fg=1;
				m_points[i].y=-m_velocities[i].y-20;
			}
			m_velocities[i]=m_points[i]-m_oldpoints[i];

			//C
			if(fg)
			{
				m_velocities[i].x*=ATTENUATION2*.95;
				m_velocities[i].z*=ATTENUATION2*.95;
			}
			else
			{
				m_velocities[i]*=ATTENUATION2;
			}
			if(i==1||i==POINTS-2)
			{
				m_velocities[i]*=ATTENUATION2;
			}
			m_oldpoints[i]=m_points[i];
		}
	}
	void Update()
	{
		for(int i=0;i<20;i++)
		{
			UpdatePhysicalSimulation();
		}
		for(int i=1;i<POINTS-1;i++)
		{
			Normals(i,m_mainnormals[i],m_subnormals[i]);
			if(m_mainnormals[i].DotProd(m_mainnormals[i-1])<0)
			{
				m_mainnormals[i]=-m_mainnormals[i];
				m_subnormals[i]=-m_subnormals[i];
			}
		}
	}

	//}EX甭郌CƌL^ĂA
	//}EX̓ɏ]ē_𓮂肷B
	//pos:C̔_
	//vec:Č
	//return:_true
	bool CrossRay(const Vector3d &pos,const Vector3d &vec)
	{
		if(!m_iscaptured)
		{
			double mindist=1000;

			m_currentpoint=-1;

			for(int i=0;i<POINTS;++i)
			{
				double d=CalcLinePointDistance(pos,vec,m_points[i]);
				if(d<ACCEPTABLE)
				{
					double d2=pos.Dist(m_points[i]);
					if(d2<mindist)
					{
						m_currentpointrecord=m_points[i];
						m_currentattracter=m_currentpointrecord;
						m_currentpoint=i;
						mindist=d2;
					}
				}
			}
			return (mindist<1000);
		}
		else
		{
			Vector3d mop=m_oldpoints[m_currentpoint];
			Vector3d mp=m_points[m_currentpoint];
			Vector3d cpr=m_currentpointrecord;//cpr:ݒڂĂ_̃hbOJn̈ʒu

			Vector3d norm=-pos;norm.Normalize();//norm:J̌Ă
			Vector3d v=vec;v.Normalize();//v:JJ[\oĂ
			Vector3d e=cpr;e-=pos;//e:posAcprւ̃xNg
			double dist=norm.DotProd(e);//mpJɐ𗎂ƂƂ̌_ƁAposƂ̋
			double dot=norm.DotProd(v);//vnorm̂Ȃp̃RTC
			v*=dist/dot;
			v+=pos;

			if(m_currentpoint!=0 && m_currentpoint!=POINTS-1)
			{
				if(v.y<-20)
					v.y=-20;
			}
			
			Vector3d vv=v;vv-=mp;
			vv*=0.5;

			if(m_currentpoint==0 || m_currentpoint==POINTS-1)
			{
				m_oldpoints[m_currentpoint]+=vv;
				m_points[m_currentpoint]+=vv;
				m_velocities[m_currentpoint]+=vv;
			}



			m_currentattracter=v;

			return true;
		}
	}

	//}EX̃hbO
	bool Capture(bool iscaptured)
	{
		if(m_currentpoint>=0)
		{
			m_iscaptured=iscaptured;
		}
		else
		{
			m_iscaptured=0;
		}
		return m_iscaptured;
	}

};


#endif

